@@ -1,6 +1,6 @@
 ;;; hexl.el --- edit a file in a hex dump format using the hexl filter.
 
-;; Copyright (C) 1989, 1994 Free Software Foundation, Inc.
+;; Copyright (C) 1989, 1994, 1998 Free Software Foundation, Inc.
 
 ;; Author: Keith Gabryelski <ag@wheaties.ai.mit.edu>
 ;; Maintainer: FSF
@@ -23,7 +23,7 @@
 ;; Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 ;; 02111-1307, USA.
 
-;;; Synched up with: FSF 19.34.
+;;; Synched up with: FSF 21.0.103.
 
 ;;; Commentary:
 
@@ -55,25 +55,52 @@
 ;; vars here
 ;;
 
-(defvar hexl-program "hexl"
+(defgroup hexl nil
+  "Edit a file in a hex dump format using the hexl filter."
+  :group 'data)
+
+
+(defcustom hexl-program "hexl"
   "The program that will hexlify and dehexlify its stdin.
 `hexl-program' will always be concatenated with `hexl-options'
-and \"-de\" when dehexlifying a buffer.")
+and \"-de\" when dehexlifying a buffer."
+  :type 'string
+  :group 'hexl)
 
-(defvar hexl-iso ""
+(defcustom hexl-iso ""
   "If your emacs can handle ISO characters, this should be set to
-\"-iso\" otherwise it should be \"\".")
+\"-iso\" otherwise it should be \"\"."
+  :type 'string
+  :group 'hexl)
 
-(defvar hexl-options (format "-hex %s" hexl-iso)
-  "Options to hexl-program that suit your needs.")
+(defcustom hexl-options (format "-hex %s" hexl-iso)
+  "Options to hexl-program that suit your needs."
+  :type 'string
+  :group 'hexl)
 
-(defvar hexlify-command
-  (format "%s%s %s" exec-directory hexl-program hexl-options)
-  "The command to use to hexlify a buffer.")
+(defcustom hexlify-command
+  (format "%s %s"
+	  (shell-quote-argument
+	   (expand-file-name hexl-program exec-directory))
+	  hexl-options)
+  "The command to use to hexlify a buffer."
+  :type 'string
+  :group 'hexl)
 
-(defvar dehexlify-command
-  (format "%s%s -de %s" exec-directory hexl-program hexl-options)
-  "The command to use to unhexlify a buffer.")
+(defcustom dehexlify-command
+  (format "%s -de %s"
+	  (shell-quote-argument
+	   (expand-file-name hexl-program exec-directory))
+	  hexl-options)
+  "The command to use to unhexlify a buffer."
+  :type 'string
+  :group 'hexl)
+
+(defcustom hexl-follow-ascii t
+  "If non-nil then highlight the ASCII character corresponding to point."
+  :type 'boolean
+  :group 'hexl
+  :version "20.3")
 
 (defvar hexl-max-address 0
   "Maximum offset into hexl buffer.")
@@ -87,12 +114,20 @@ and \"-de\" when dehexlifying a buffer.")
 (defvar hexl-mode-old-require-final-newline)
 (defvar hexl-mode-old-syntax-table)
 
+(defvar hexl-ascii-extent nil
+  "Extent used to highlight ASCII element corresponding to current point.")
+(make-variable-buffer-local 'hexl-ascii-extent)
+
 ;; routines
 
+(put 'hexl-mode 'mode-class 'special)
+
 ;;;###autoload
 (defun hexl-mode (&optional arg)
-  "\\<hexl-mode-map>
-A major mode for editing binary files in hex dump format.
+  "\\<hexl-mode-map>A mode for editing binary files in hex dump format.
+This is not an ordinary major mode; it alters some aspects
+if the current mode's behavior, but not all; also, you can exit
+Hexl mode and return to the previous mode using `hexl-mode-exit'.
 
 This function automatically converts a buffer into the hexl format
 using the function `hexlify-buffer'.
@@ -158,13 +193,11 @@ into the buffer at the current point.
 Note: saving the file with any of the usual Emacs commands
 will actually convert it back to binary format while saving.
 
-You can use \\[hexl-find-file] to visit a file in hexl-mode.
+You can use \\[hexl-find-file] to visit a file in Hexl mode.
 
 \\[describe-bindings] for advanced commands."
   (interactive "p")
-  (if (eq major-mode 'hexl-mode)
-      (error "You are already in hexl mode")
-
+  (unless (eq major-mode 'hexl-mode)
     (let ((modified (buffer-modified-p))
 	  (inhibit-read-only t)
 	  (original-point (1- (point)))
@@ -214,10 +247,13 @@ You can use \\[hexl-find-file] to visit a file in hexl-mode.
     (add-hook 'after-revert-hook 'hexl-after-revert-hook nil t)
 
     (make-local-hook 'change-major-mode-hook)
-    (add-hook 'change-major-mode-hook 'hexl-maybe-dehexlify-buffer nil t))
+    (add-hook 'change-major-mode-hook 'hexl-maybe-dehexlify-buffer nil t)
+
+    (if hexl-follow-ascii (hexl-follow-ascii 1)))
   (run-hooks 'hexl-mode-hook))
 
 (defun hexl-after-revert-hook ()
+  (setq hexl-max-address (1- (buffer-size)))
   (hexlify-buffer)
   (set-buffer-modified-p nil))
 
@@ -231,7 +267,7 @@ You can use \\[hexl-find-file] to visit a file in hexl-mode.
 			       (save-excursion
 				 (let ((buf (generate-new-buffer " hexl"))
 				       (name (buffer-name))
-				       (file-name (buffer-file-name))
+				       ;(file-name (buffer-file-name))
 				       (start (point-min))
 				       (end (point-max))
 				       modified)
@@ -240,8 +276,7 @@ You can use \\[hexl-find-file] to visit a file in hexl-mode.
 				   (set-buffer name)
 				   (dehexlify-buffer)
 				   ;; Prevent infinite recursion.
-				   (let ((hexl-in-save-buffer t)
-					 (buffer-file-type t)) ; for ms-dos
+				   (let ((hexl-in-save-buffer t))
 				     (save-buffer))
 				   (setq modified (buffer-modified-p))
 				   (delete-region (point-min) (point-max))
@@ -277,6 +312,8 @@ With arg, don't unhexlify buffer."
 
   (remove-hook 'after-revert-hook 'hexl-after-revert-hook t)
   (remove-hook 'change-major-mode-hook 'hexl-maybe-dehexlify-buffer t)
+  (remove-hook 'post-command-hook 'hexl-follow-ascii-find t)
+  (setq hexl-ascii-extent nil)
 
   (setq write-contents-hooks hexl-mode-old-write-contents-hooks)
   (setq require-final-newline hexl-mode-old-require-final-newline)
@@ -302,7 +339,7 @@ Ask the user for confirmation."
 (defun hexl-current-address (&optional validate)
   "Return current hexl-address."
   (interactive)
-  (let ((current-column (- (% (point) 68) 11)) 
+  (let ((current-column (- (% (point) 68) 11))
 	(hexl-address 0))
     (if (< current-column 0)
 	(if validate
@@ -316,7 +353,7 @@ Ask the user for confirmation."
     hexl-address))
 
 (defun hexl-address-to-marker (address)
-  "Return marker for ADDRESS."
+  "Return buffer position for ADDRESS."
   (interactive "nAddress: ")
   (+ (* (/ address 16) 68) 11 (/ (* (% address 16) 5) 2)))
 
@@ -325,7 +362,7 @@ Ask the user for confirmation."
 Signal error if ADDRESS out of range."
   (interactive "nAddress: ")
   (if (or (< address 0) (> address hexl-max-address))
-	  (error "Out of hexl region."))
+	  (error "Out of hexl region"))
   (goto-char (hexl-address-to-marker address)))
 
 (defun hexl-goto-hex-address (hex-address)
@@ -507,18 +544,30 @@ With prefix arg N, puts point N bytes of the way from the true beginning."
   (hexl-scroll-up (- arg)))
 
 (defun hexl-scroll-up (arg)
-  "Scroll hexl buffer window upward ARG lines; or near full window if no ARG."
+  "Scroll hexl buffer window upward ARG lines; or near full window if no ARG.
+If there's no byte at the target address, move to the first or last line."
   (interactive "P")
   (if (null arg)
       (setq arg (1- (window-height)))
     (setq arg (prefix-numeric-value arg)))
-  (let ((movement (* arg 16))
-	(address (hexl-current-address)))
-    (if (or (> (+ address movement) hexl-max-address)
-	    (< (+ address movement) 0))
-	(message "Out of hexl region.")
-      (hexl-goto-address (+ address movement))
-      (recenter 0))))
+  (let* ((movement (* arg 16))
+	 (address (hexl-current-address))
+	 (dest (+ address movement)))
+    (cond
+     ;; If possible, try to stay at the same offset from the beginning
+     ;; of the 16-byte group, even if we move to the first or last
+     ;; group.
+     ((and (> dest hexl-max-address)
+	   (>= (% hexl-max-address 16) (% address 16)))
+      (setq dest (+ (logand hexl-max-address -16) (% address 16))))
+     ((> dest hexl-max-address)
+      (setq dest hexl-max-address))
+     ((< dest 0)
+      (setq dest (% address 16))))
+    (if (/= dest (+ address movement))
+	(message "Out of hexl region."))
+    (hexl-goto-address dest)
+    (recenter 0)))
 
 (defun hexl-beginning-of-1k-page ()
   "Go to beginning of 1k boundary."
@@ -564,10 +613,16 @@ This discards the buffer's undo information."
        (or (y-or-n-p "Converting to hexl format discards undo info; ok? ")
 	   (error "Aborted")))
   (setq buffer-undo-list nil)
-  (let ((binary-process-output nil) ; for Ms-Dos
-	(binary-process-input t)
+  ;; Don't decode text in the ASCII part of `hexl' program output.
+  (let ((coding-system-for-read 'raw-text)
+	(coding-system-for-write 
+	 (or (and (boundp 'buffer-file-coding-system)
+		  buffer-file-coding-system)
+	     'raw-text))
 	(buffer-undo-list t))
-    (shell-command-on-region (point-min) (point-max) hexlify-command t)))
+    (shell-command-on-region (point-min) (point-max) hexlify-command t)
+    (if (> (point) (hexl-address-to-marker hexl-max-address))
+	(hexl-goto-address hexl-max-address))))
 
 (defun dehexlify-buffer ()
   "Convert a hexl format buffer to binary.
@@ -577,8 +632,11 @@ This discards the buffer's undo information."
        (or (y-or-n-p "Converting from hexl format discards undo info; ok? ")
 	   (error "Aborted")))
   (setq buffer-undo-list nil)
-  (let ((binary-process-output t) ; for Ms-Dos
-	(binary-process-input nil)
+  (let ((coding-system-for-write 'raw-text)
+	(coding-system-for-read 
+	 (or (and (boundp 'buffer-file-coding-system)
+		  buffer-file-coding-system)
+	     'raw-text))
 	(buffer-undo-list t))
     (shell-command-on-region (point-min) (point-max) dehexlify-command t)))
 
@@ -599,13 +657,13 @@ This discards the buffer's undo information."
     (let ((ch (logior character 32)))
       (if (and (>= ch ?a) (<= ch ?f))
 	  (- ch (- ?a 10))
-	(error "Invalid hex digit `%c'." ch)))))
+	(error "Invalid hex digit `%c'" ch)))))
 
 (defun hexl-oct-char-to-integer (character)
   "Take a char and return its value as if it was a octal digit."
   (if (and (>= character ?0) (<= character ?7))
       (- character ?0)
-    (error "Invalid octal digit `%c'." character)))
+    (error "Invalid octal digit `%c'" character)))
 
 (defun hexl-printable-character (ch)
   "Return a displayable string for character CH."
@@ -659,15 +717,40 @@ This discards the buffer's undo information."
   (interactive "p")
   (let ((num (hexl-hex-string-to-integer (read-string "Hex number: "))))
     (if (or (> num 255) (< num 0))
-	(error "Hex number out of range.")
+	(error "Hex number out of range")
       (hexl-insert-char num arg))))
 
+(defun hexl-insert-hex-string (str arg)
+  "Insert hexadecimal string STR at point ARG times.
+Embedded whitespace, dashes, and periods in the string are ignored."
+  (interactive "sHex string: \np")
+  (setq str (replace-in-string str "[- \t.]" ""))
+  (let ((chars '()))
+    (let ((len (length str))
+	  (idx 0))
+      (if (eq (logand len 1) 1)
+	  (let ((num (hexl-hex-string-to-integer (substring str 0 1))))
+	    (setq chars (cons num chars))
+	    (setq idx 1)))
+      (while (< idx len)
+	(let* ((nidx (+ idx 2))
+	       (num (hexl-hex-string-to-integer (substring str idx nidx))))
+	  (setq chars (cons num chars))
+	  (setq idx nidx))))
+    (setq chars (nreverse chars))
+    (while (> arg 0)
+      (let ((chars chars))
+	(while chars
+	  (hexl-insert-char (car chars) 1)
+	  (setq chars (cdr chars))))
+      (setq arg (- arg 1)))))
+
 (defun hexl-insert-decimal-char (arg)
   "Insert a ASCII char ARG times at point for a given decimal number."
   (interactive "p")
   (let ((num (string-to-int (read-string "Decimal Number: "))))
     (if (or (> num 255) (< num 0))
-	(error "Decimal number out of range.")
+	(error "Decimal number out of range")
       (hexl-insert-char num arg))))
 
 (defun hexl-insert-octal-char (arg)
@@ -675,9 +758,47 @@ This discards the buffer's undo information."
   (interactive "p")
   (let ((num (hexl-octal-string-to-integer (read-string "Octal Number: "))))
     (if (or (> num 255) (< num 0))
-	(error "Decimal number out of range.")
+	(error "Decimal number out of range")
       (hexl-insert-char num arg))))
 
+(defun hexl-follow-ascii (&optional arg)
+  "Toggle following ASCII in Hexl buffers.
+With prefix ARG, turn on following if and only if ARG is positive.
+When following is enabled, the ASCII character corresponding to the
+element under the point is highlighted.
+Customize the variable `hexl-follow-ascii' to disable this feature."
+  (interactive "P")
+  (let ((on-p (if arg 
+		  (> (prefix-numeric-value arg) 0)
+	       (not hexl-ascii-extent))))
+
+    (make-local-hook 'post-command-hook)
+		    
+    (if on-p
+      ;; turn it on
+      (if (not hexl-ascii-extent)
+	  (progn
+	    (setq hexl-ascii-extent (make-extent 1 1)
+		  hexl-follow-ascii t)
+	    (set-extent-property hexl-ascii-extent 'face 'highlight)
+	    (add-hook 'post-command-hook 'hexl-follow-ascii-find nil t)))
+      ;; turn it off
+      (if hexl-ascii-extent
+	  (progn
+	    (detach-extent hexl-ascii-extent)
+	    (setq hexl-ascii-extent nil
+		  hexl-follow-ascii nil)
+	    (remove-hook 'post-command-hook 'hexl-follow-ascii-find t)
+	    )))))
+
+(defun hexl-follow-ascii-find ()
+  "Find and highlight the ASCII element corresponding to current point."
+  (let ((pos (+ 51
+		(- (point) (current-column))
+		(mod (hexl-current-address) 16))))
+    (set-extent-endpoints hexl-ascii-extent pos (1+ pos))
+    ))
+
 ;; startup stuff.
 
 (if hexl-mode-map
@@ -692,7 +813,10 @@ This discards the buffer's undo information."
   (define-key hexl-mode-map [(meta right)] 'hexl-forward-short)
   (define-key hexl-mode-map [next] 'hexl-scroll-up)
   (define-key hexl-mode-map [prior] 'hexl-scroll-down)
-  (define-key hexl-mode-map [home] 'hexl-beginning-of-buffer)
+  (define-key hexl-mode-map [home] 'hexl-beginning-of-line)
+  (define-key hexl-mode-map [end] 'hexl-end-of-line)
+  (define-key hexl-mode-map [(control home)] 'hexl-beginning-of-buffer)
+  (define-key hexl-mode-map [(control end)] 'hexl-end-of-buffer)
   (define-key hexl-mode-map [deletechar] 'undefined)
   (define-key hexl-mode-map [deleteline] 'undefined)
   (define-key hexl-mode-map [insertline] 'undefined)
@@ -730,48 +854,26 @@ This discards the buffer's undo information."
 
   (define-key hexl-mode-map "\e\C-a" 'hexl-beginning-of-512b-page)
   (define-key hexl-mode-map "\e\C-b" 'hexl-backward-short)
-  (define-key hexl-mode-map "\e\C-c" 'undefined)
   (define-key hexl-mode-map "\e\C-d" 'hexl-insert-decimal-char)
   (define-key hexl-mode-map "\e\C-e" 'hexl-end-of-512b-page)
   (define-key hexl-mode-map "\e\C-f" 'hexl-forward-short)
-  (define-key hexl-mode-map "\e\C-g" 'undefined)
-  (define-key hexl-mode-map "\e\C-h" 'undefined)
   (define-key hexl-mode-map "\e\C-i" 'undefined)
   (define-key hexl-mode-map "\e\C-j" 'undefined)
   (define-key hexl-mode-map "\e\C-k" 'undefined)
-  (define-key hexl-mode-map "\e\C-l" 'undefined)
-  (define-key hexl-mode-map "\e\C-m" 'undefined)
-  (define-key hexl-mode-map "\e\C-n" 'undefined)
   (define-key hexl-mode-map "\e\C-o" 'hexl-insert-octal-char)
-  (define-key hexl-mode-map "\e\C-p" 'undefined)
   (define-key hexl-mode-map "\e\C-q" 'undefined)
-  (define-key hexl-mode-map "\e\C-r" 'undefined)
-  (define-key hexl-mode-map "\e\C-s" 'undefined)
   (define-key hexl-mode-map "\e\C-t" 'undefined)
-  (define-key hexl-mode-map "\e\C-u" 'undefined)
-
-  (define-key hexl-mode-map "\e\C-w" 'undefined)
   (define-key hexl-mode-map "\e\C-x" 'hexl-insert-hex-char)
-  (define-key hexl-mode-map "\e\C-y" 'undefined)
-
-  (define-key hexl-mode-map "\ea" 'undefined)
   (define-key hexl-mode-map "\eb" 'hexl-backward-word)
   (define-key hexl-mode-map "\ec" 'undefined)
   (define-key hexl-mode-map "\ed" 'undefined)
-  (define-key hexl-mode-map "\ee" 'undefined)
   (define-key hexl-mode-map "\ef" 'hexl-forward-word)
   (define-key hexl-mode-map "\eg" 'hexl-goto-hex-address)
-  (define-key hexl-mode-map "\eh" 'undefined)
   (define-key hexl-mode-map "\ei" 'undefined)
   (define-key hexl-mode-map "\ej" 'hexl-goto-address)
   (define-key hexl-mode-map "\ek" 'undefined)
   (define-key hexl-mode-map "\el" 'undefined)
-  (define-key hexl-mode-map "\em" 'undefined)
-  (define-key hexl-mode-map "\en" 'undefined)
-  (define-key hexl-mode-map "\eo" 'undefined)
-  (define-key hexl-mode-map "\ep" 'undefined)
   (define-key hexl-mode-map "\eq" 'undefined)
-  (define-key hexl-mode-map "\er" 'undefined)
   (define-key hexl-mode-map "\es" 'undefined)
   (define-key hexl-mode-map "\et" 'undefined)
   (define-key hexl-mode-map "\eu" 'undefined)
@@ -789,4 +891,6 @@ This discards the buffer's undo information."
   (define-key hexl-mode-map "\C-x\C-s" 'hexl-save-buffer)
   (define-key hexl-mode-map "\C-x\C-t" 'undefined))
 
+(provide 'hexl)
+
 ;;; hexl.el ends here
